"""
TrafficNode module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Guille guille.gutierrezmorote@concordia.ca
"""

from typing import List, TypeVar

from hub.city_model_structure.attributes.edge import Edge
from hub.city_model_structure.attributes.node import Node
from hub.city_model_structure.attributes.point import Point

Connection = TypeVar('Connection')
TrafficEdge = TypeVar('TrafficEdge')


class TrafficNode(Node):
  """
  TrafficNode class
  """

  def __init__(self, name, coordinates, node_type='TrafficNode', edges=None, prohibitions=None, connections=None):
    super().__init__(name, edges)
    if connections is None:
      connections = []
    if prohibitions is None:
      prohibitions = []
    self._coordinates = coordinates
    self._prohibitions = prohibitions
    self._connections = connections
    self._node_type = node_type

  @property
  def node_type(self):
    """
    Get node type
    :return: str
    """
    return self._node_type

  @property
  def coordinates(self) -> Point:
    """
    Get node coordinates
    :return: Point
    """
    return self._coordinates

  @coordinates.setter
  def coordinates(self, value):
    """
    Set node coordinates
    :param value: Point
    """
    self._coordinates = value

  @property
  def edges(self) -> List[TrafficEdge]:
    """
    get edges delimited by the node
    :return: [TrafficEdge]
    """
    return self._edges

  @property
  def prohibitions(self) -> [(Edge, Edge)]:
    """
    Get node prohibitions
    :return: [(Edge, Edge)]
    """
    return self._prohibitions

  @prohibitions.setter
  def prohibitions(self, value):
    """
    Set node prohibitions
    :param value: [(Edge, Edge)]
    """
    self._prohibitions = value

  @property
  def connections(self) -> List[Connection]:
    """
    Get node connections
    :return: [Connection]
    """
    return self._connections

  @connections.setter
  def connections(self, value):
    """
    Set node connections
    :param value: [Connection]
    """
    self._connections = value



"""
Greenery catalog data model Plant class
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from hub.catalog_factories.data_models.greenery.soil import Soil as hub_soil


class Plant:
  """
  Plant class
  """
  def __init__(self, category, plant):
    self._name = plant.name
    self._category = category
    self._height = plant.height
    self._leaf_area_index = plant.leafAreaIndex
    self._leaf_reflectivity = plant.leafReflectivity
    self._leaf_emissivity = plant.leafEmissivity
    self._minimal_stomatal_resistance = plant.minimalStomatalResistance
    self._co2_sequestration = plant.co2Sequestration
    self._grows_on = []
    for soil in plant.growsOn:
      self._grows_on.append(hub_soil(soil))

  @property
  def name(self):
    """
    Get plant name
    :return: string
    """
    return self._name

  @property
  def category(self):
    """
    Get plant category name
    :return: string
    """
    return self._category

  @property
  def height(self):
    """
    Get plant height in m
    :return: float
    """
    return self._height

  @property
  def leaf_area_index(self):
    """
    Get plant leaf area index
    :return: float
    """
    return self._leaf_area_index

  @property
  def leaf_reflectivity(self):
    """
    Get plant leaf area index
    :return: float
    """
    return self._leaf_reflectivity

  @property
  def leaf_emissivity(self):
    """
    Get plant leaf emissivity
    :return: float
    """
    return self._leaf_emissivity

  @property
  def minimal_stomatal_resistance(self):
    """
    Get plant minimal stomatal resistance in s/m
    :return: float
    """
    return self._minimal_stomatal_resistance

  @property
  def co2_sequestration(self):
    """
    Get plant co2 sequestration capacity in kg CO2 equivalent
    :return: float
    """
    return self._co2_sequestration

  @property
  def grows_on(self) -> [hub_soil]:
    """
    Get plant compatible soils
    :return: [Soil]
    """
    return self._grows_on

  def to_dictionary(self):
    """Class content to dictionary"""
    _soils = []
    for _soil in self.grows_on:
      _soils.append(_soil.to_dictionary())
    content = {'Plant': {'name': self.name,
                         'category': self.category,
                         'height [m]': self.height,
                         'leaf area index': self.leaf_area_index,
                         'leaf reflectivity': self.leaf_reflectivity,
                         'leaf emissivity': self.leaf_emissivity,
                         'minimal stomatal resistance [s/m]': self.minimal_stomatal_resistance,
                         'co2 sequestration [kg????]': self.co2_sequestration,
                         'soils where it grows on': _soils
                         }
               }

    return content



"""
Lane module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import List, Union


class Lane:
  """
  Lane class
  """

  def __init__(self):
    self._index = None
    self._allow = None
    self._disallow = None
    self._change_left = None
    self._change_right = None
    self._speed = None
    self._width = None

  @property
  def index(self) -> Union[None, int]:
    """
    Get lane index
    The enumeration index of the lane (0 is the rightmost lane, <NUMBER_LANES>-1 is the leftmost one)
    :return: None or int
    """
    return self._index

  @index.setter
  def index(self, value):
    """
    Set lane index
    The enumeration index of the lane (0 is the rightmost lane, <NUMBER_LANES>-1 is the leftmost one)
    :param value: int
    """
    if value is not None:
      self._index = int(value)

  @property
  def allow(self) -> Union[None, List[str]]:
    """
    Get the list of allowed vehicle classes
    :return: None or [str]
    """
    return self._allow

  @allow.setter
  def allow(self, value):
    """
    Set the list of allowed vehicle classes setter
    :param value: [str]
    """
    if value is not None:
      self._allow = [str(i) for i in value]

  @property
  def disallow(self) -> Union[None, List[str]]:
    """
    Get the list of not allowed vehicle classes
    :return: None or [str]
    """
    return self._disallow

  @disallow.setter
  def disallow(self, value):
    """
    Get the list of not allowed vehicle classes setter
    :param value: [str]
    """
    if value is not None:
      self._disallow = [str(i) for i in value]

  @property
  def change_left(self) -> Union[None, List[str]]:
    """
    Get the list of vehicle classes that may change left from this lane
    :return: None or [str]
    """
    return self._change_left

  @change_left.setter
  def change_left(self, value):
    """
    Set the list of vehicle classes that may change left from this lane
    :param value: [str]
    """
    if value is not None:
      self._change_left = [str(i) for i in value]

  @property
  def change_right(self) -> Union[None, List[str]]:
    """
    Get the list of vehicle classes that may change right from this lane
    :return: None or [str]
    """
    return self._change_right

  @change_right.setter
  def change_right(self, value):
    """
    Set the list of vehicle classes that may change right from this lane
    :param value: [str]
    """
    if value is not None:
      self._change_right = [str(i) for i in value]

  @property
  def speed(self) -> Union[None, float]:
    """
    Get the lane speed in m/s
    :return: None or float
    """
    return self._speed

  @speed.setter
  def speed(self, value):
    """
    Set the lane speed in m/s
    :param value: float
    """
    if value is not None:
      self._speed = float(value)

  @property
  def width(self) -> Union[None, float]:
    """
    Get the lane width in meters
    :return: None or float
    """
    return self._width

  @width.setter
  def width(self, value):
    """
    Set the lane width in meters
    :param value: float
    """
    if value is not None:
      self._width = float(value)



"""
CityObjectsCluster module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from abc import ABC
from typing import List
from hub.city_model_structure.iot.sensor import Sensor
from hub.city_model_structure.city_object import CityObject


class CityObjectsCluster(ABC, CityObject):
  """
  CityObjectsCluster(ABC) class
  """
  def __init__(self, name, cluster_type, city_objects):
    self._name = name
    self._cluster_type = cluster_type
    self._city_objects = city_objects
    self._sensors = []
    super().__init__(name, None)

  @property
  def name(self):
    """
    Get cluster name
    :return: str
    """
    return self._name

  @property
  def type(self):
    """
    raises not implemented error
    """
    raise NotImplementedError

  @property
  def city_objects(self):
    """
    raises not implemented error
    """
    raise NotImplementedError

  def add_city_object(self, city_object) -> List[CityObject]:
    """
    add new object to the cluster
    :return: [CityObjects]
    """
    if self._city_objects is None:
      self._city_objects = [city_object]
    else:
      self._city_objects.append(city_object)
    return self._city_objects

  @property
  def sensors(self) -> List[Sensor]:
    """
    Get sensors belonging to the city objects cluster
    :return: [Sensor]
    """
    return self._sensors

  @sensors.setter
  def sensors(self, value):
    """
    Set sensors belonging to the city objects cluster
    :param value: [Sensor]
    """
    self._sensors = value



"""
Usage catalog usage
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
from typing import Union

from hub.catalog_factories.data_models.usages.appliances import Appliances
from hub.catalog_factories.data_models.usages.lighting import Lighting
from hub.catalog_factories.data_models.usages.occupancy import Occupancy
from hub.catalog_factories.data_models.usages.thermal_control import ThermalControl
from hub.catalog_factories.data_models.usages.domestic_hot_water import DomesticHotWater


class Usage:
  """
  Usage class
  """
  def __init__(self, name,
               hours_day,
               days_year,
               mechanical_air_change,
               ventilation_rate,
               occupancy,
               lighting,
               appliances,
               thermal_control,
               domestic_hot_water):
    self._name = name
    self._hours_day = hours_day
    self._days_year = days_year
    self._mechanical_air_change = mechanical_air_change
    self._ventilation_rate = ventilation_rate
    self._occupancy = occupancy
    self._lighting = lighting
    self._appliances = appliances
    self._thermal_control = thermal_control
    self._domestic_hot_water = domestic_hot_water

  @property
  def name(self) -> Union[None, str]:
    """
    Get usage zone usage name
    :return: None or str
    """
    return self._name

  @property
  def hours_day(self) -> Union[None, float]:
    """
    Get usage zone usage hours per day
    :return: None or float
    """
    return self._hours_day

  @property
  def days_year(self) -> Union[None, float]:
    """
    Get usage zone usage days per year
    :return: None or float
    """
    return self._days_year

  @property
  def mechanical_air_change(self) -> Union[None, float]:
    """
    Get usage zone mechanical air change in air change per hour (ACH)
    :return: None or float
    """
    return self._mechanical_air_change

  @property
  def ventilation_rate(self) -> Union[None, float]:
    """
    Get usage zone ventilation rate in m3/m2s
    :return: None or float
    """
    return self._ventilation_rate

  @property
  def occupancy(self) -> Union[None, Occupancy]:
    """
    Get occupancy in the usage zone
    :return: None or Occupancy
    """
    return self._occupancy

  @property
  def lighting(self) -> Union[None, Lighting]:
    """
    Get lighting information
    :return: None or Lighting
    """
    return self._lighting

  @lighting.setter
  def lighting(self, value):
    """
    Set lighting information
    :param value: Lighting
    """
    self._lighting = value

  @property
  def appliances(self) -> Union[None, Appliances]:
    """
    Get appliances information
    :return: None or Appliances
    """
    return self._appliances

  @property
  def thermal_control(self) -> Union[None, ThermalControl]:
    """
    Get thermal control information
    :return: None or ThermalControl
    """
    return self._thermal_control

  @property
  def domestic_hot_water(self) -> Union[None, DomesticHotWater]:
    """
    Get domestic hot water information
    :return: None or DomesticHotWater
    """
    return self._domestic_hot_water

  def to_dictionary(self):
    """Class content to dictionary"""
    content = {'Usage': {'name': self.name,
                         'hours a day': self.hours_day,
                         'days a year': self.days_year,
                         'mechanical air change [ACH]': self.mechanical_air_change,
                         'ventilation rate [m3/sm2]': self.ventilation_rate,
                         'occupancy': self.occupancy.to_dictionary(),
                         'lighting': self.lighting.to_dictionary(),
                         'appliances': self.appliances.to_dictionary(),
                         'thermal control': self.thermal_control.to_dictionary(),
                         'domestic hot water': self.domestic_hot_water.to_dictionary()
                         }
               }
    return content



"""
Bus edge module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import List, TypeVar
from hub.city_model_structure.attributes.edge import Edge

BusNode = TypeVar('BusNode')


class BusEdge(Edge):
  """
  BusEdge class
  Each edge is unidirectional and starts at the "from" node and ends at the "to" node
  """

  def __init__(self, name, nodes, edge_type='BusEdge'):
    super().__init__(name, nodes)
    self._edge_type = edge_type
    self._average_travel_time = None

  @property
  def edge_type(self):
    """
    Get the edge type
    :return: str
    """
    return self._edge_type

  @property
  def nodes(self) -> List[BusNode]:
    """
    Get delimiting nodes for the edge
    :return: [BusNode]
    """
    return self._nodes

  @property
  def average_travel_time(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._average_travel_time



"""
GeometryFactory retrieve the specific geometric module to load the given format
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
import geopandas

from hub.city_model_structure.city import City
from hub.helpers.utils import validate_import_export_type
from hub.imports.geometry.citygml import CityGml
from hub.imports.geometry.geojson import Geojson
from hub.imports.geometry.gpandas import GPandas
from hub.imports.geometry.obj import Obj
from hub.imports.geometry.rhino import Rhino


class GeometryFactory:
  """
  GeometryFactory class
  """
  def __init__(self, file_type,
               path=None,
               data_frame=None,
               aliases_field=None,
               height_field=None,
               year_of_construction_field=None,
               function_field=None,
               function_to_hub=None):
    self._file_type = '_' + file_type.lower()
    validate_import_export_type(GeometryFactory, file_type)
    self._path = path
    self._data_frame = data_frame
    self._aliases_field = aliases_field
    self._height_field = height_field
    self._year_of_construction_field = year_of_construction_field
    self._function_field = function_field
    self._function_to_hub = function_to_hub

  @property
  def _citygml(self) -> City:
    """
    Enrich the city by using CityGML information as data source
    :return: City
    """
    return CityGml(self._path,
                   self._height_field,
                   self._year_of_construction_field,
                   self._function_field,
                   self._function_to_hub).city

  @property
  def _obj(self) -> City:
    """
    Enrich the city by using OBJ information as data source
    :return: City
    """
    return Obj(self._path).city

  @property
  def _gpandas(self) -> City:
    """
    Enrich the city by using GeoPandas information as data source
    :return: City
    """
    if self._data_frame is None:
      self._data_frame = geopandas.read_file(self._path)
    return GPandas(self._data_frame).city

  @property
  def _geojson(self) -> City:
    """
    Enrich the city by using Geojson information as data source
    :return: City
    """
    return Geojson(self._path,
                   self._aliases_field,
                   self._height_field,
                   self._year_of_construction_field,
                   self._function_field,
                   self._function_to_hub).city

  @property
  def _rhino(self) -> City:
    """
    Enrich the city by using Rhino information as data source
    :return: City
    """
    return Rhino(self._path).city

  @property
  def city(self) -> City:
    """
    Enrich the city given to the class using the class given handler
    :return: City
    """
    return getattr(self, self._file_type, lambda: None)



"""
Insel Heap pump energy demand and fossil fuel consumption
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder peter.yefi@gmail.cm
"""

from pathlib import Path
import pandas as pd


class InselHeatPumpEnergyDemand:
  """
  Import Energy demand and fossil fuel consumption results
  """
  def __init__(self, city, base_path, hp_model):
    """
    :param city: the city
    :param base_path: the insel simulation output file
    :param hp_model: the heatpump model for both air source and water to water
    """
    self._city = city
    self._hp_model = hp_model
    with open(Path(base_path).resolve(), 'r', encoding='utf8') as csv_file:
      df = pd.read_csv(csv_file)
    self._monthly_electricity_demand = df.iloc[:, 1]
    self._monthly_fossil_fuel_consumption = df.iloc[:, 2]

  def enrich(self):
    """
    Enrich the city with the heat pump information
    """
    for energy_system in self._city.energy_systems:
      if energy_system.air_source_hp is not None:
        if energy_system.air_source_hp.model == self._hp_model:
          energy_system.air_source_hp.hp_monthly_fossil_consumption = self._monthly_fossil_fuel_consumption

      if energy_system.water_to_water_hp is not None:
        if energy_system.water_to_water_hp.model == self._hp_model:
          energy_system.water_to_water_hp.hp_monthly_electricity_demand = self._monthly_electricity_demand



"""
export a city into Obj format
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from pathlib import Path


class Obj:
  """
  Export to obj format
  """
  def __init__(self, city, path):
    self._city = city
    self._path = path
    self._export()

  def _to_vertex(self, coordinate):
    x = coordinate[0] - self._city.lower_corner[0]
    y = coordinate[1] - self._city.lower_corner[1]
    z = coordinate[2] - self._city.lower_corner[2]
    return f'v {x} {y} {z}\n'

  def _export(self):
    if self._city.name is None:
      self._city.name = 'unknown_city'
    file_name = self._city.name + '.obj'
    file_path = (Path(self._path).resolve() / file_name).resolve()
    vertices = {}
    with open(file_path, 'w', encoding='utf-8') as obj:
      obj.write("# cerc-hub export\n")
      vertex_index = 0
      faces = []
      for building in self._city.buildings:
        obj.write(f'# building {building.name}\n')
        obj.write(f'g {building.name}\n')
        obj.write('s off\n')
        for surface in building.surfaces:
          obj.write(f'# surface {surface.name}\n')
          face = 'f '
          for coordinate in surface.perimeter_polygon.coordinates:
            vertex = self._to_vertex(coordinate)
            if vertex not in vertices:
              vertex_index += 1
              vertices[vertex] = vertex_index
              current = vertex_index
              obj.write(vertex)
            else:
              current = vertices[vertex]

            face = f'{face} {current}'

          faces.append(f'{face} {face.split(" ")[1]}\n')
          obj.writelines(faces)
          faces = []



"""
AirSourceHeatPumpParameters import the heat pump information
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.comCode
contributor Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
import io

import itertools
from typing import List
from typing import Dict
import pandas as pd
import numpy as np
from scipy.optimize import curve_fit
from hub.city_model_structure.energy_systems.air_source_hp import AirSourceHP
from hub.city_model_structure.energy_system import EnergySystem


class AirSourceHeatPumpParameters:
  """
  AirSourceHeatPumpParameters class
  """

  def __init__(self, city, base_path):
    self._city = city
    self._base_path = (base_path / 'heat_pumps/air_source.xlsx').resolve()

  def _read_file(self) -> Dict:
    """
    reads xlsx file containing the heat pump information
    into a dictionary
    :return : Dict
    """
    with open(self._base_path, 'rb') as xls:
      xl_file = pd.read_excel(io.BytesIO(xls.read()), sheet_name=None)

    cooling_data = {}
    heating_data = {}

    for sheet, _ in xl_file.items():
      if 'Summary' in sheet:
        continue

      # Remove nan rows and columns and extract cooling and heating data
      # for each sheet
      df = xl_file[sheet].dropna(axis=1, how='all')
      cooling_df = df.iloc[4:34, 0:8]
      heating_df = df.iloc[4:29, 8:20]

      # extract the data into dictionaries each sheet is a key entry in the
      # dictionary
      cooling_data[sheet] = {}
      heating_data[sheet] = {}
      i = 0
      # for each sheet extract data for twout/Ta.RU temperatures. Thus, the twout
      # temp is the key for the values of pf,pa,qw data
      while i < 25:
        cooling_data[sheet][cooling_df.iloc[i][0]] = cooling_df.iloc[i + 1:i + 4, 2:8].values.tolist()
        heating_data[sheet][heating_df.iloc[i][0]] = heating_df.iloc[i + 1:i + 4, 2:8].values.tolist()
        i = i + 5
      # extract the last cooling data
      cooling_data[sheet][cooling_df.iloc[i][0]] = cooling_df.iloc[i + 1:i + 4, 2:8].values.tolist()
    return {"cooling": cooling_data, "heating": heating_data}

  def enrich_city(self):
    """
    Enriches the city with information from file
    """
    heat_pump_data = self._read_file()
    for (k_cool, v_cool), (_, v_heat) in zip(heat_pump_data["cooling"].items(), heat_pump_data["heating"].items()):
      heat_pump = AirSourceHP()
      heat_pump.model = k_cool
      h_data = self._extract_heat_pump_data(v_heat)
      c_data = self._extract_heat_pump_data(v_cool)
      heat_pump.cooling_capacity = c_data[0]
      heat_pump.cooling_comp_power = c_data[1]
      heat_pump.cooling_capacity_coff = self._compute_coefficients(c_data, "cool")
      heat_pump.heating_capacity = h_data[0]
      heat_pump.heating_comp_power = h_data[1]
      heat_pump.heating_capacity_coff = self._compute_coefficients(h_data)

      energy_system = EnergySystem(f'{heat_pump.model} capacity heat pump', [])
      energy_system.air_source_hp = heat_pump
      self._city.add_city_object(energy_system)
    return self._city

  @staticmethod
  def _extract_heat_pump_data(heat_pump_capacity_data: Dict) -> [List, List]:
    """
    Fetches a list of metric based data for heat pump for various temperature,
    e.g. cooling capacity data for 12 capacity heat pump
    for 6,7,8,9,10 and 11 degree Celsius
    :param heat_pump_capacity_data: the heat pump capacity data from the
    which the metric specific data is fetched: {List}
    :return: List
    """
    cooling_heating_capacity_data = []
    compressor_power_data = []
    for _, metric_data in heat_pump_capacity_data.items():
      cooling_heating_capacity_data.append(metric_data[0])
      compressor_power_data.append(metric_data[1])
    return [cooling_heating_capacity_data, compressor_power_data]

  def _compute_coefficients(self, heat_pump_data: List, data_type="heat") -> List[float]:
    """
    Compute heat output and electrical demand coefficients
    from heating and cooling performance data
    :param heat_pump_data: a list of heat pump data. e.g. cooling capacity
    :param data_type: string to indicate if data is cooling performance data
    or heating performance data
    :return: Tuple[Dict, Dict]
    """
    # Determine the recurrence of temperature values. 6 repetitions for
    # cooling performance and 5 repetition for heating performance
    temp_multiplier = 5 if data_type == "heat" else 6
    out_temp = [25, 30, 32, 35, 40, 45] * temp_multiplier

    heat_x_values = np.repeat([-5, 0, 7, 10, 15], 6)
    cool_x_values = np.repeat([6, 7, 8, 9, 10, 11], 6)
    x_values = heat_x_values if data_type == "heat" else cool_x_values
    x_values = x_values.tolist()
    # convert list of lists to one list

    hp_data = [i / j for i, j in
               zip(list(itertools.chain.from_iterable(heat_pump_data[0])),
                   list(itertools.chain.from_iterable(heat_pump_data[1])))]

    # Compute heat output coefficients
    popt, _ = curve_fit(self._objective_function, [x_values, out_temp], hp_data)
    return popt.tolist()

  @staticmethod
  def _objective_function(xdata: List, a1: float, a2: float, a3: float, a4: float, a5: float, a6: float) -> float:
    """
    Objective function for computing coefficients
    :param xdata:
    :param a1: float
    :param a2: float
    :param a3: float
    :param a4: float
    :param a5: float
    :param a6: float
    :return:
    """
    x, y = xdata
    return (a1 * x ** 2) + (a2 * x) + (a3 * x * y) + (a4 * y) + (a5 * y ** 2) + a6



"""
Origin-Destination network module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import List
from hub.city_model_structure.network import Network
from hub.city_model_structure.transport.origin_destination_edge import OriginDestinationEdge
from hub.city_model_structure.transport.origin_destination_node import OriginDestinationNode


class OriginDestinationNetwork(Network):
  """
  OriginDestinationNetwork(Network) class
  """
  def __init__(self, name, edges=None, nodes=None):
    super().__init__(name, edges, nodes)
    self._type = "OriginDestinationNetwork"

  @property
  def type(self):
    """
    Get network type
    :return: str
    """
    return self._type

  @property
  def edges(self) -> List[OriginDestinationEdge]:
    """
    Get network edges
    :return: [OriginDestinationEdge]
    """
    return self._edges

  @property
  def nodes(self) -> List[OriginDestinationNode]:
    """
    Get network nodes
    :return: [OriginDestinationNode]
    """
    return self._nodes



"""
NrcanUsageParameters extracts the usage properties from NRCAN catalog and assigns to each building
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import logging

import hub.helpers.constants as cte
from hub.helpers.dictionaries import Dictionaries
from hub.city_model_structure.building_demand.usage import Usage
from hub.city_model_structure.building_demand.lighting import Lighting
from hub.city_model_structure.building_demand.occupancy import Occupancy
from hub.city_model_structure.building_demand.appliances import Appliances
from hub.city_model_structure.building_demand.thermal_control import ThermalControl
from hub.city_model_structure.building_demand.domestic_hot_water import DomesticHotWater
from hub.catalog_factories.usage_catalog_factory import UsageCatalogFactory


class NrcanUsageParameters:
  """
  NrcanUsageParameters class
  """
  def __init__(self, city):
    self._city = city

  def enrich_buildings(self):
    """
    Returns the city with the usage parameters assigned to the buildings
    :return:
    """
    city = self._city
    nrcan_catalog = UsageCatalogFactory('nrcan').catalog
    comnet_catalog = UsageCatalogFactory('comnet').catalog

    for building in city.buildings:
      usage_name = Dictionaries().hub_usage_to_nrcan_usage[building.function]
      try:
        archetype_usage = self._search_archetypes(nrcan_catalog, usage_name)
      except KeyError:
        logging.error('Building %s has unknown usage archetype for usage %s', building.name, usage_name)
        continue

      comnet_usage_name = Dictionaries().hub_usage_to_comnet_usage[building.function]
      try:
        comnet_archetype_usage = self._search_archetypes(comnet_catalog, comnet_usage_name)
      except KeyError:
        logging.error('Building %s has unknown usage archetype for usage %s', building.name, comnet_usage_name)
        continue

      for internal_zone in building.internal_zones:
        if len(building.internal_zones) > 1:
          volume_per_area = 0
          if internal_zone.area is None:
            logging.error('Building %s has internal zone area not defined, ACH cannot be calculated for usage %s',
                          building.name, usage_name)
            continue
          if internal_zone.volume is None:
            logging.error('Building %s has internal zone volume not defined, ACH cannot be calculated for usage %s',
                          building.name, usage_name)
            continue
          if internal_zone.area <= 0:
            logging.error('Building %s has internal zone area equal to 0, ACH cannot be calculated for usage %s',
                          building.name,  usage_name)
            continue
          volume_per_area += internal_zone.volume / internal_zone.area
        else:
          if building.storeys_above_ground is None:
            logging.error('Building %s no number of storeys assigned, ACH cannot be calculated for usage %s',
                          building.name,  usage_name)
            continue
          volume_per_area = building.volume / building.floor_area / building.storeys_above_ground

        usage = Usage()
        usage.name = usage_name
        self._assign_values(usage, archetype_usage, volume_per_area, building.cold_water_temperature)
        self._assign_comnet_extra_values(usage, comnet_archetype_usage, archetype_usage.occupancy.occupancy_density)
        usage.percentage = 1
        self._calculate_reduced_values_from_extended_library(usage, archetype_usage)

        internal_zone.usages = [usage]

  @staticmethod
  def _search_archetypes(catalog, usage_name):
    archetypes = catalog.entries('archetypes').usages
    for building_archetype in archetypes:
      if str(usage_name) == str(building_archetype.name):
        return building_archetype
    raise KeyError('archetype not found')

  @staticmethod
  def _assign_values(usage, archetype, volume_per_area, cold_water_temperature):
    if archetype.mechanical_air_change > 0:
      # ACH
      usage.mechanical_air_change = archetype.mechanical_air_change
    elif archetype.ventilation_rate > 0:
      # m3/m2.s to ACH
      usage.mechanical_air_change = archetype.ventilation_rate / volume_per_area * cte.HOUR_TO_SECONDS
    else:
      usage.mechanical_air_change = 0
    _occupancy = Occupancy()
    _occupancy.occupancy_density = archetype.occupancy.occupancy_density
    _occupancy.sensible_radiative_internal_gain = archetype.occupancy.sensible_radiative_internal_gain
    _occupancy.latent_internal_gain = archetype.occupancy.latent_internal_gain
    _occupancy.sensible_convective_internal_gain = archetype.occupancy.sensible_convective_internal_gain
    _occupancy.occupancy_schedules = archetype.occupancy.schedules
    usage.occupancy = _occupancy
    _lighting = Lighting()
    _lighting.density = archetype.lighting.density
    _lighting.convective_fraction = archetype.lighting.convective_fraction
    _lighting.radiative_fraction = archetype.lighting.radiative_fraction
    _lighting.latent_fraction = archetype.lighting.latent_fraction
    _lighting.schedules = archetype.lighting.schedules
    usage.lighting = _lighting
    _appliances = Appliances()
    _appliances.density = archetype.appliances.density
    _appliances.convective_fraction = archetype.appliances.convective_fraction
    _appliances.radiative_fraction = archetype.appliances.radiative_fraction
    _appliances.latent_fraction = archetype.appliances.latent_fraction
    _appliances.schedules = archetype.appliances.schedules
    usage.appliances = _appliances
    _control = ThermalControl()
    _control.cooling_set_point_schedules = archetype.thermal_control.cooling_set_point_schedules
    _control.heating_set_point_schedules = archetype.thermal_control.heating_set_point_schedules
    _control.hvac_availability_schedules = archetype.thermal_control.hvac_availability_schedules
    usage.thermal_control = _control
    _domestic_hot_water = DomesticHotWater()
    _domestic_hot_water.peak_flow = archetype.domestic_hot_water.peak_flow
    _domestic_hot_water.service_temperature = archetype.domestic_hot_water.service_temperature
    density = None
    if len(cold_water_temperature) > 0:
      cold_temperature = cold_water_temperature[cte.YEAR]['epw']
      density = (
          archetype.domestic_hot_water.peak_flow * cte.WATER_DENSITY * cte.WATER_HEAT_CAPACITY *
          (archetype.domestic_hot_water.service_temperature - cold_temperature)
      )
    _domestic_hot_water.density = density
    _domestic_hot_water.schedules = archetype.domestic_hot_water.schedules
    usage.domestic_hot_water = _domestic_hot_water

  @staticmethod
  def _assign_comnet_extra_values(usage, archetype, occupancy_density):
    _occupancy = usage.occupancy
    archetype_density = archetype.occupancy.occupancy_density
    if archetype_density == 0:
      _occupancy.sensible_radiative_internal_gain = 0
      _occupancy.latent_internal_gain = 0
      _occupancy.sensible_convective_internal_gain = 0
    else:
      _occupancy.sensible_radiative_internal_gain = archetype.occupancy.sensible_radiative_internal_gain \
                                                    * occupancy_density / archetype_density
      _occupancy.latent_internal_gain = archetype.occupancy.latent_internal_gain * occupancy_density / archetype_density
      _occupancy.sensible_convective_internal_gain = (
          archetype.occupancy.sensible_convective_internal_gain * occupancy_density / archetype_density
      )

  @staticmethod
  def _calculate_reduced_values_from_extended_library(usage, archetype):
    number_of_days_per_type = {'WD': 251, 'Sat': 52, 'Sun': 62}
    total = 0
    for schedule in archetype.thermal_control.hvac_availability_schedules:
      if schedule.day_types[0] == cte.SATURDAY:
        for value in schedule.values:
          total += value * number_of_days_per_type['Sat']
      elif schedule.day_types[0] == cte.SUNDAY:
        for value in schedule.values:
          total += value * number_of_days_per_type['Sun']
      else:
        for value in schedule.values:
          total += value * number_of_days_per_type['WD']

    usage.hours_day = total / 365
    usage.days_year = 365

    max_heating_setpoint = cte.MIN_FLOAT
    min_heating_setpoint = cte.MAX_FLOAT

    for schedule in archetype.thermal_control.heating_set_point_schedules:
      if schedule.values is None:
        max_heating_setpoint = None
        min_heating_setpoint = None
        break
      if max(schedule.values) > max_heating_setpoint:
        max_heating_setpoint = max(schedule.values)
      if min(schedule.values) < min_heating_setpoint:
        min_heating_setpoint = min(schedule.values)

    min_cooling_setpoint = cte.MAX_FLOAT
    for schedule in archetype.thermal_control.cooling_set_point_schedules:
      if schedule.values is None:
        min_cooling_setpoint = None
        break
      if min(schedule.values) < min_cooling_setpoint:
        min_cooling_setpoint = min(schedule.values)

    usage.thermal_control.mean_heating_set_point = max_heating_setpoint
    usage.thermal_control.heating_set_back = min_heating_setpoint
    usage.thermal_control.mean_cooling_set_point = min_cooling_setpoint



"""
Traffic edge module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Guille guille.gutierrezmorote@concordia.ca
"""

from typing import List, Union
from hub.city_model_structure.attributes.edge import Edge
from hub.city_model_structure.transport.traffic_node import TrafficNode
from hub.city_model_structure.transport.lane import Lane


class TrafficEdge(Edge):
  """
  TrafficEdge class
  Each edge is unidirectional and starts at the "from" node and ends at the "to" node
  """

  def __init__(self, name, nodes, priority, speed, lanes, length, allows=None, disallows=None, sidewalk_width=None,
               edge_type='TrafficEdge'):
    super().__init__(name, nodes)
    self._edge_type = edge_type
    self._lanes = lanes
    self._priority = priority
    self._speed = speed
    self._length = length
    self._allows = allows
    self._disallows = disallows
    self._sidewalk_width = sidewalk_width

  @property
  def edge_type(self):
    """
    Get the edge type
    :return: str
    """
    return self._edge_type

  @property
  def nodes(self) -> List[TrafficNode]:
    """
    Get delimiting nodes for the edge
    :return: [TrafficNode]
    """
    return self._nodes

  @property
  def lanes(self) -> List[Lane]:
    """
    Get the lanes on an edge
    :return: List[Lane]
    """
    return self._lanes

  @lanes.setter
  def lanes(self, value):
    """
    Set the lanes on an edge
    :param value: List[Lane]
    """
    self._lanes = value

  @property
  def priority(self) -> Union[None, int]:
    """
    Get the priority between different road types.
    It starts with one; higher numbers represent more important roads.
    :return: None or int
    """
    return self._priority

  @priority.setter
  def priority(self, value):
    """
    Set the priority between different road types.
    It starts with one; higher numbers represent more important roads.
    :param value: int
    """
    if value is not None:
      self._priority = int(value)

  @property
  def speed(self) -> Union[None, float]:
    """
    Get he speed limit in m/s
    :return: None or float
    """
    return self._speed

  @speed.setter
  def speed(self, value):
    """
    Set the speed limit in m/s
    :param value: float
    """
    if value is not None:
      self._speed = float(value)

  @property
  def length(self) -> Union[None, float]:
    """
    Get the lane length in meters
    :return: None or float
    """
    return self._length

  @length.setter
  def length(self, value):
    """
    Set the lane length in meters
    :param value: float
    """
    if value is not None:
      self._length = float(value)

  @property
  def allows(self) -> Union[None, List[str]]:
    """
    Get the list of allowed vehicle classes
    :return: None or [str]
    """
    return self._allows

  @allows.setter
  def allows(self, value):
    """
    Set the list of allowed vehicle classes
    :param value: [str]
    """
    if value is not None:
      self._allows = [str(i) for i in value]

  @property
  def disallows(self) -> Union[None, List[str]]:
    """
    Get the list of not allowed vehicle classes
    :return: None or [str]
    """
    return self._disallows

  @disallows.setter
  def disallows(self, value):
    """
    Set the list of not allowed vehicle classes
    :param value: [str]
    """
    if value is not None:
      self._disallows = [str(i) for i in value]



"""
Incomes included in the costs catalog
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import Union


class Income:
  """
  Income class
  """
  def __init__(self, construction_subsidy=None,
               hvac_subsidy=None,
               photovoltaic_subsidy=None,
               electricity_export=None,
               reductions_tax=None):

    self._construction_subsidy = construction_subsidy
    self._hvac_subsidy = hvac_subsidy
    self._photovoltaic_subsidy = photovoltaic_subsidy
    self._electricity_export = electricity_export
    self._reductions_tax = reductions_tax

  @property
  def construction_subsidy(self) -> Union[None, float]:
    """
    Get subsidy for construction in percentage
    :return: None or float
    """
    return self._construction_subsidy

  @property
  def hvac_subsidy(self) -> Union[None, float]:
    """
    Get subsidy for HVAC system in percentage
    :return: None or float
    """
    return self._hvac_subsidy

  @property
  def photovoltaic_subsidy(self) -> Union[None, float]:
    """
    Get subsidy PV systems in percentage
    :return: None or float
    """
    return self._photovoltaic_subsidy

  @property
  def electricity_export(self) -> Union[None, float]:
    """
    Get electricity export incomes in currency per J
    :return: None or float
    """
    return self._electricity_export

  @property
  def reductions_tax(self) -> Union[None, float]:
    """
    Get reduction in taxes in percentage (-)
    :return: None or float
    """
    return self._reductions_tax

  def to_dictionary(self):
    """Class content to dictionary"""
    content = {'Income': {'construction subsidy [%]': self.construction_subsidy,
                          'hvac subsidy [%]': self.hvac_subsidy,
                          'photovoltaic subsidy [%]': self.photovoltaic_subsidy,
                          'electricity export [currency/J]': self.electricity_export,
                          'reductions tax': self.reductions_tax
                          }
               }

    return content



"""
Dictionaries module for hub usage to Eilat usage
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import hub.helpers.constants as cte


class HubUsageToEilatUsage:
  """
  Hub usage to Eilat usage class
  """

  def __init__(self):
    self._dictionary = {
      cte.RESIDENTIAL: 'Residential',
      cte.HOTEL: 'Hotel employees',
      cte.DORMITORY: 'Dormitory'
    }

  @property
  def dictionary(self) -> dict:
    """
    Get the dictionary
    :return: {}
    """
    return self._dictionary



"""
weather helper
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
import logging
import math
import calendar as cal
import pandas as pd
import numpy as np
import hub.helpers.constants as cte


class Weather:
  """
  Weather class
  """

  _epw_file = {
    'CA.02.5935': 'https://energyplus-weather.s3.amazonaws.com/north_and_central_america_wmo_region_4/CAN/BC/CAN_BC_Summerland.717680_CWEC/CAN_BC_Summerland.717680_CWEC.epw',
    'CA.10.06': 'https://energyplus-weather.s3.amazonaws.com/north_and_central_america_wmo_region_4/CAN/PQ/CAN_PQ_Montreal.Intl.AP.716270_CWEC/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw',
    'CA.10.13': 'https://energyplus-weather.s3.amazonaws.com/north_and_central_america_wmo_region_4/CAN/PQ/CAN_PQ_Montreal.Intl.AP.716270_CWEC/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw',
    'CA.10.14': 'https://energyplus-weather.s3.amazonaws.com/north_and_central_america_wmo_region_4/CAN/PQ/CAN_PQ_Montreal.Intl.AP.716270_CWEC/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw',
    'CA.10.16': 'https://energyplus-weather.s3.amazonaws.com/north_and_central_america_wmo_region_4/CAN/PQ/CAN_PQ_Montreal.Intl.AP.716270_CWEC/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw',
    'DE.01.082': 'https://energyplus-weather.s3.amazonaws.com/europe_wmo_region_6/DEU/DEU_Stuttgart.107380_IWEC/DEU_Stuttgart.107380_IWEC.epw',
    'US.NY.047': 'https://energyplus-weather.s3.amazonaws.com/north_and_central_america_wmo_region_4/USA/NY/USA_NY_New.York.City-Central.Park.94728_TMY/USA_NY_New.York.City-Central.Park.94728_TMY.epw',
    'CA.10.12': 'https://energyplus-weather.s3.amazonaws.com/north_and_central_america_wmo_region_4/CAN/PQ/CAN_PQ_Quebec.717140_CWEC/CAN_PQ_Quebec.717140_CWEC.epw',
    'IL.01.': 'https://energyplus-weather.s3.amazonaws.com/europe_wmo_region_6/ISR/ISR_Eilat.401990_MSI/ISR_Eilat.401990_MSI.epw'
  }
  # todo: this dictionary need to be completed, a data science student task?

  @staticmethod
  def sky_temperature(ambient_temperature):
    """
    Get sky temperature from ambient temperature in Celsius
    :return: List[float]
    """
    # Swinbank - Source sky model approximation(1963) based on cloudiness statistics(32 %) in the United States
    # ambient temperatures( in °C)
    # sky temperatures( in °C)
    values = []
    for temperature in ambient_temperature:
      value = 0.037536 * math.pow((temperature + cte.KELVIN), 1.5) \
              + 0.32 * (temperature + cte.KELVIN) - cte.KELVIN
      values.append(value)
    return values

  @staticmethod
  def cold_water_temperature(ambient_temperature):
    """
    Get cold water temperature from ambient temperature in Celsius
    :return: dict
    """
    # Equation from "TOWARDS DEVELOPMENT OF AN ALGORITHM FOR MAINS WATER TEMPERATURE", 2004, Jay Burch
    # and Craig Christensen, National Renewable Energy Laboratory
    # ambient temperatures( in °C)
    # cold water temperatures( in °C)
    ambient_temperature_fahrenheit = []
    average_temperature = 0
    maximum_temperature = -1000
    minimum_temperature = 1000
    for temperature in ambient_temperature:
      value = temperature * 9 / 5 + 32
      ambient_temperature_fahrenheit.append(value)
      average_temperature += value / 8760
      if value > maximum_temperature:
        maximum_temperature = value
      if value < minimum_temperature:
        minimum_temperature = value
    delta_temperature = maximum_temperature - minimum_temperature
    ratio = 0.4 + 0.01 * (average_temperature - 44)
    lag = 35 - 1 * (average_temperature - 44)
    cold_temperature = []
    for temperature in ambient_temperature_fahrenheit:
      radians = (0.986 * (temperature-15-lag) - 90) * math.pi / 180
      cold_temperature.append((average_temperature + 6 + ratio * (delta_temperature/2) * math.sin(radians) - 32) * 5/9)
    return pd.DataFrame(cold_temperature, columns=['epw'])

  def get_monthly_mean_values(self, values):
    """
    Get the monthly mean for the given values
    :return: float
    """
    out = None
    if values is not None:
      if 'month' not in values.columns:
        values = pd.concat([self.month_hour, pd.DataFrame(values)], axis=1)
      out = values.groupby('month', as_index=False).mean()
      del out['month']
    return out

  @staticmethod
  def get_yearly_mean_values(values):
    """
    Get the yearly mean for the given values
    :return: float
    """
    return values.mean()

  def get_total_month(self, values):
    """
    Get the total value the given values
    :return: float
    """
    out = None
    if values is not None:
      if 'month' not in values.columns:
        values = pd.concat([self.month_hour, pd.DataFrame(values)], axis=1)
      out = pd.DataFrame(values).groupby('month', as_index=False).sum()
      del out['month']
    return out

  @property
  def month_hour(self):
    """
    returns a DataFrame that has x values of the month number (January = 1, February = 2...),
    being x the number of hours of the corresponding month
    :return: DataFrame(int)
    """
    array = []
    for i in range(0, 12):
      days_of_month = cal.monthrange(2015, i+1)[1]
      total_hours = days_of_month * 24
      array = np.concatenate((array, np.full(total_hours, i + 1)))
    return pd.DataFrame(array, columns=['month'])

  def epw_file(self, region_code):
    """
    returns the url for the weather file for the given location or default (Montreal data)
    :return: str
    """
    if region_code not in self._epw_file:
      logging.warning('Specific weather data unknown for %s using Montreal data instead', region_code)
      return self._epw_file['CA.10.06']
    return self._epw_file[region_code]



"""
Test greenery factory test and validate the greenery construction
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from unittest import TestCase

from hub.catalog_factories.greenery_catalog_factory import GreeneryCatalogFactory


class TestGreeneryCatalog(TestCase):
  def test_catalog(self):
    catalog = GreeneryCatalogFactory('nrel').catalog
    catalog_categories = catalog.names()
    vegetations = catalog.names('vegetations')
    plants = catalog.names('plants')
    soils = catalog.names('soils')
    self.assertTrue(len(catalog_categories) == 3)
    self.assertTrue(len(vegetations['vegetations']) == 4)
    self.assertTrue(len(plants['plants']) == 14)
    self.assertTrue(len(soils['soils']) == 6)
    with self.assertRaises(ValueError):
      catalog.names('unknown')

    # retrieving all the entries should not raise any exceptions
    for category in catalog_categories:
      for value in catalog_categories[category]:
        catalog.get_entry(value)

    with self.assertRaises(IndexError):
      catalog.get_entry('unknown')

    self.assertTrue(len(catalog.entries().vegetations) == 4)
    self.assertTrue(len(catalog.entries().plants) == 14)
    self.assertTrue(len(catalog.entries().soils) == 6)



"""
Layers module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
import uuid
from typing import Union
from hub.city_model_structure.building_demand.material import Material


class Layer:
  """
  Layer class
  """
  def __init__(self):
    self._material = None
    self._thickness = None
    self._id = None

  @property
  def id(self):
    """
    Get layer id, a universally unique identifier randomly generated
    :return: str
    """
    if self._id is None:
      self._id = uuid.uuid4()
    return self._id

  @property
  def material(self) -> Material:
    """
    Get layer material
    :return: Material
    """
    return self._material

  @material.setter
  def material(self, value):
    """
    Set layer material
    :param value: Material
    """
    self._material = value

  @property
  def thickness(self) -> Union[None, float]:
    """
    Get layer thickness in meters
    :return: None or float
    """
    return self._thickness

  @thickness.setter
  def thickness(self, value):
    """
    Get layer thickness in meters
    :param value: float
    """
    if value is not None:
      self._thickness = float(value)



"""
CityGml module parses citygml_classes files and import the geometry into the city model structure
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

import numpy as np
import xmltodict

from hub.city_model_structure.building import Building
from hub.city_model_structure.city import City
from hub.city_model_structure.parts_consisting_building import PartsConsistingBuilding
from hub.imports.geometry.citygml_classes.citygml_lod1 import CityGmlLod1
from hub.imports.geometry.citygml_classes.citygml_lod2 import CityGmlLod2


class CityGml:
  """
  CityGml class
  """

  def __init__(self,
               path,
               extrusion_height_field=None,
               year_of_construction_field=None,
               function_field=None,
               function_to_hub=None):
    self._city = None
    self._lod = None
    self._lod1_tags = ['lod1Solid', 'lod1MultiSurface']
    self._lod2_tags = ['lod2Solid', 'lod2MultiSurface', 'lod2MultiCurve']
    self._extrusion_height_field = extrusion_height_field
    self._function_to_hub = function_to_hub
    if function_field is None:
      function_field = 'function'
    if year_of_construction_field is None:
      year_of_construction_field = 'yearOfConstruction'
    self._year_of_construction_field = year_of_construction_field
    self._function_field = function_field

    self._lower_corner = None
    self._upper_corner = None
    with open(path, 'r', encoding='utf8') as gml:
      # Clean the namespaces is an important task to prevent wrong ns:field due poor citygml_classes implementations
      force_list = ('cityObjectMember', 'curveMember', 'boundedBy', 'surfaceMember', 'consistsOfBuildingPart')
      self._gml = xmltodict.parse(gml.read(), process_namespaces=True, xml_attribs=True, namespaces={
        'http://www.opengis.net/gml': None,
        'http://www.opengis.net/citygml/1.0': None,
        'http://www.opengis.net/citygml/building/1.0': None,
        'http://schemas.opengis.net/citygml/building/1.0/building.xsd': None,
        'http://www.opengis.net/citygml/appearance/1.0': None,
        'http://schemas.opengis.net/citygml/appearance/1.0/appearance.xsd': None,
        'http://www.opengis.net/citygml/relief/1.0': None,
        'http://schemas.opengis.net/citygml/relief/1.0/relief.xsd': None,
        'http://www.opengis.net/citygml/generics/1.0': None,
        'http://www.w3.org/2001/XMLSchema-instance': None,
        'urn:oasis:names:tc:ciq:xsdschema:xAL:2.0': None,
        'http://www.w3.org/1999/xlink': None,
        'http://www.opengis.net/citygml/relief/2.0': None,
        'http://www.opengis.net/citygml/building/2.0': None,
        'http://www.opengis.net/citygml/building/2.0 http://schemas.opengis.net/citygml/building/2.0/building.xsd '
        'http://www.opengis.net/citygml/relief/2.0 http://schemas.opengis.net/citygml/relief/2.0/relief.xsd" '
        'xmlns="http://www.opengis.net/citygml/2.0': None,
        'http://www.opengis.net/citygml/2.0': None
      }, force_list=force_list)

      self._city_objects = None
      if 'boundedBy' in self._gml['CityModel']:
        for bound in self._gml['CityModel']['boundedBy']:
          envelope = bound['Envelope']
          if '#text' in envelope['lowerCorner']:
            self._lower_corner = np.fromstring(envelope['lowerCorner']['#text'], dtype=float, sep=' ')
            self._upper_corner = np.fromstring(envelope['upperCorner']['#text'], dtype=float, sep=' ')
          else:
            self._lower_corner = np.fromstring(envelope['lowerCorner'], dtype=float, sep=' ')
            self._upper_corner = np.fromstring(envelope['upperCorner'], dtype=float, sep=' ')
          if '@srsName' in envelope:
            self._srs_name = envelope['@srsName']
          else:
            # If not coordinate system given assuming hub standard
            self._srs_name = "EPSG:26911"
      else:
        # get the boundary from the city objects instead
        for city_object_member in self._gml['CityModel']['cityObjectMember']:
          city_object = city_object_member['Building']
          if 'boundedBy' in city_object:
            for bound in city_object['boundedBy']:
              if 'Envelope' not in bound:
                continue
              envelope = bound['Envelope']
              self._srs_name = envelope['@srsName']
              if '#text' in envelope['lowerCorner']:
                lower_corner = np.fromstring(envelope['lowerCorner']['#text'], dtype=float, sep=' ')
                upper_corner = np.fromstring(envelope['upperCorner']['#text'], dtype=float, sep=' ')
              else:
                lower_corner = np.fromstring(envelope['lowerCorner'], dtype=float, sep=' ')
                upper_corner = np.fromstring(envelope['upperCorner'], dtype=float, sep=' ')
              if self._lower_corner is None:
                self._lower_corner = lower_corner
                self._upper_corner = upper_corner
              else:
                self._lower_corner[0] = min(self._lower_corner[0], lower_corner[0])
                self._lower_corner[1] = min(self._lower_corner[1], lower_corner[1])
                self._lower_corner[2] = min(self._lower_corner[2], lower_corner[2])
                self._upper_corner[0] = max(self._upper_corner[0], upper_corner[0])
                self._upper_corner[1] = max(self._upper_corner[1], upper_corner[1])
                self._upper_corner[2] = max(self._upper_corner[2], upper_corner[2])

  @property
  def content(self):
    """
    Get cityGml raw content
    :return: str
    """
    return self._gml

  def _create_building(self, city_object):
    name = city_object['@id']
    function = None
    year_of_construction = None
    if self._year_of_construction_field in city_object:
      year_of_construction = city_object[self._year_of_construction_field]
    if self._function_field in city_object:
      function = city_object[self._function_field]
      if not isinstance(function, str):
        function = function['#text']
      if self._function_to_hub is not None:
        # use the transformation dictionary to retrieve the proper function
        function = self._function_to_hub[function]

    if any(key in city_object for key in self._lod1_tags):
      if self._lod is None or self._lod > 1:
        self._lod = 1
      surfaces = CityGmlLod1(city_object).surfaces
    elif any(key in city_object for key in self._lod2_tags):
      if self._lod is None or self._lod > 2:
        self._lod = 2
      surfaces = CityGmlLod2(city_object).surfaces
    else:
      raise NotImplementedError("Not supported level of detail")
    return Building(name, surfaces, year_of_construction, function, terrains=None)

  def _create_parts_consisting_building(self, city_object):
    name = city_object['@id']
    building_parts = []
    for part in city_object['consistsOfBuildingPart']:
      building = self._create_building(part['BuildingPart'])
      self._city.add_city_object(building)
      building_parts.append(building)
    return PartsConsistingBuilding(name, building_parts)

  @property
  def city(self) -> City:
    """
    Get city model structure enriched with the geometry information
    :return: City
    """

    if self._city is None:
      self._city = City(self._lower_corner, self._upper_corner, self._srs_name)
      for city_object_member in self._gml['CityModel']['cityObjectMember']:
        city_object = city_object_member['Building']
        if 'consistsOfBuildingPart' in city_object:
          self._city.add_city_objects_cluster(self._create_parts_consisting_building(city_object))
        else:
          self._city.add_city_object(self._create_building(city_object))
      self._city.level_of_detail.geometry = self._lod
      for building in self._city.buildings:
        building.level_of_detail.geometry = self._lod
    return self._city



"""
Polygon module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from __future__ import annotations

import logging
import math
import sys
from typing import List

import numpy as np
import trimesh.creation
import trimesh.geometry
import trimesh.intersections
from shapely.geometry.polygon import Polygon as shapley_polygon
from trimesh import Trimesh

from hub.city_model_structure.attributes.plane import Plane
from hub.city_model_structure.attributes.point import Point


class Polygon:
  """
  Polygon class
  """
  def __init__(self, coordinates):
    self._area = None
    self._points = None
    self._points_list = None
    self._normal = None
    self._inverse = None
    self._edges = None
    self._coordinates = coordinates
    self._triangles = None
    self._vertices = None
    self._faces = None
    self._plane = None

  @property
  def points(self) -> List[Point]:
    """
    Get the points belonging to the polygon [[x, y, z],...]
    :return: [Point]
    """
    if self._points is None:
      self._points = []
      for coordinate in self.coordinates:
        self._points.append(Point(coordinate))
    return self._points

  @property
  def plane(self) -> Plane:
    """
    Get the polygon plane
    :return: Plane
    """
    if self._plane is None:
      self._plane = Plane(normal=self.normal, origin=self.points[0])
    return self._plane

  @property
  def coordinates(self) -> List[np.ndarray]:
    """
    Get the points in the shape of its coordinates belonging to the polygon [[x, y, z],...]
    :return: [np.ndarray]
    """
    return self._coordinates

  @property
  def points_list(self) -> np.ndarray:
    """
    Get the solid surface point coordinates list [x, y, z, x, y, z,...]
    :return: np.ndarray
    """
    if self._points_list is None:
      s = self.coordinates
      self._points_list = np.reshape(s, len(s) * 3)
    return self._points_list

  @property
  def edges(self) -> List[List[Point]]:
    """
    Get polygon edges list
    :return: [[Point]]
    """
    if self._edges is None:
      self._edges = []
      for i in range(0, len(self.points) - 1):
        point_1 = self.points[i]
        point_2 = self.points[i + 1]
        self._edges.append([point_1, point_2])
      self._edges.append([self.points[len(self.points) - 1], self.points[0]])
    return self._edges

  @property
  def area(self):
    """
    Get surface area in square meters
    :return: float
    """
    if self._area is None:
      self._area = 0
      for triangle in self.triangles:
        a_b = np.zeros(3)
        a_c = np.zeros(3)
        for i in range(0, 3):
          a_b[i] = triangle.coordinates[1][i] - triangle.coordinates[0][i]
          a_c[i] = triangle.coordinates[2][i] - triangle.coordinates[0][i]
        self._area += np.linalg.norm(np.cross(a_b, a_c)) / 2
    return self._area

  @area.setter
  def area(self, value):
    self._area = value

  @property
  def normal(self) -> np.ndarray:
    """
    Get surface normal vector
    :return: np.ndarray
    """
    if self._normal is None:
      points = self.coordinates
      # todo: IF THE FIRST ONE IS 0, START WITH THE NEXT
      point_origin = points[len(points) - 2]
      vector_1 = points[len(points) - 1] - point_origin
      vector_2 = points[0] - point_origin
      vector_3 = points[1] - point_origin
      cross_product = np.cross(vector_1, vector_2)
      if np.linalg.norm(cross_product) != 0:
        cross_product = cross_product / np.linalg.norm(cross_product)
        alpha = self._angle_between_vectors(vector_1, vector_2)
      else:
        cross_product = [0, 0, 0]
        alpha = 0
      if len(points) == 3:
        return cross_product
      if np.linalg.norm(cross_product) == 0:
        return cross_product
      alpha += self._angle(vector_2, vector_3, cross_product)
      for i in range(0, len(points) - 4):
        vector_1 = points[i + 1] - point_origin
        vector_2 = points[i + 2] - point_origin
        alpha += self._angle(vector_1, vector_2, cross_product)
      vector_1 = points[len(points) - 1] - point_origin
      vector_2 = points[0] - point_origin
      if alpha < 0:
        cross_product = np.cross(vector_2, vector_1)
      else:
        cross_product = np.cross(vector_1, vector_2)
      self._normal = cross_product / np.linalg.norm(cross_product)
    return self._normal

  @staticmethod
  def _angle(vector_1, vector_2, cross_product):
    """
    alpha angle in radians
    :param vector_1: [float]
    :param vector_2: [float]
    :param cross_product: [float]
    :return: float
    """
    accepted_normal_difference = 0.01
    cross_product_next = np.cross(vector_1, vector_2)
    if np.linalg.norm(cross_product_next) != 0:
      cross_product_next = cross_product_next / np.linalg.norm(cross_product_next)
      alpha = Polygon._angle_between_vectors(vector_1, vector_2)
    else:
      cross_product_next = [0, 0, 0]
      alpha = 0
    delta_normals = 0
    for j in range(0, 3):
      delta_normals += cross_product[j] - cross_product_next[j]
    if np.abs(delta_normals) < accepted_normal_difference:
      return alpha
    return -alpha

  @staticmethod
  def triangle_mesh(vertices, normal) -> Trimesh:
    """
    Get the triangulated mesh for the polygon
    :return: Trimesh
    """
    min_x = 1e16
    min_y = 1e16
    min_z = 1e16
    for vertex in vertices:
      if vertex[0] < min_x:
        min_x = vertex[0]
      if vertex[1] < min_y:
        min_y = vertex[1]
      if vertex[2] < min_z:
        min_z = vertex[2]

    new_vertices = []
    for vertex in vertices:
      vertex = [vertex[0]-min_x, vertex[1]-min_y, vertex[2]-min_z]
      new_vertices.append(vertex)

    transformation_matrix = trimesh.geometry.plane_transform(origin=new_vertices[0], normal=normal)

    coordinates = []
    for vertex in vertices:
      transformed_vertex = [vertex[0]-min_x, vertex[1]-min_y, vertex[2]-min_z, 1]
      transformed_vertex = np.dot(transformation_matrix, transformed_vertex)
      coordinate = [transformed_vertex[0], transformed_vertex[1]]
      coordinates.append(coordinate)

    polygon = shapley_polygon(coordinates)

    try:
      _, faces = trimesh.creation.triangulate_polygon(polygon, engine='triangle')

      mesh = Trimesh(vertices=vertices, faces=faces)

      # check orientation
      normal_sum = 0
      for i in range(0, 3):
        normal_sum += normal[i] + mesh.face_normals[0][i]

      if abs(normal_sum) <= 1E-10:
        new_faces = []
        for face in faces:
          new_face = []
          for i in range(0, len(face)):
            new_face.append(face[len(face)-i-1])
          new_faces.append(new_face)
        mesh = Trimesh(vertices=vertices, faces=new_faces)
      return mesh

    except ValueError:
      logging.error('Not able to triangulate polygon\n')
      _vertices = [[0, 0, 0], [0, 0, 1], [0, 1, 0]]
      _faces = [[0, 1, 2]]
      return Trimesh(vertices=_vertices, faces=_faces)

  @property
  def triangles(self) -> List[Polygon]:
    """
    Triangulate the polygon and return a list of triangular polygons
    :return: [Polygon]
    """
    if self._triangles is None:
      self._triangles = []
      _mesh = self.triangle_mesh(self.coordinates, self.normal)
      for face in _mesh.faces:
        points = []
        for vertex in face:
          points.append(self.coordinates[vertex])
        polygon = Polygon(points)
        self._triangles.append(polygon)
    return self._triangles

  @staticmethod
  def _angle_between_vectors(vec_1, vec_2):
    """
    angle between vectors in radians
    :param vec_1: vector
    :param vec_2: vector
    :return: float
    """
    if np.linalg.norm(vec_1) == 0 or np.linalg.norm(vec_2) == 0:
      sys.stderr.write("Warning: impossible to calculate angle between planes' normal. Return 0\n")
      return 0
    cosine = np.dot(vec_1, vec_2) / np.linalg.norm(vec_1) / np.linalg.norm(vec_2)
    if cosine > 1 and cosine - 1 < 1e-5:
      cosine = 1
    elif cosine < -1 and cosine + 1 > -1e-5:
      cosine = -1
    alpha = math.acos(cosine)
    return alpha

  @property
  def inverse(self):
    """
    Get the polygon coordinates in reversed order
    :return: [np.ndarray]
    """
    if self._inverse is None:
      self._inverse = self.coordinates[::-1]
    return self._inverse

  def divide(self, plane):
    """
    Divides the polygon in two by a plane
    :param plane: plane that intersects with self to divide it in two parts (Plane)
    :return: Polygon, Polygon, [Point]
    """
    tri_polygons = Trimesh(vertices=self.vertices, faces=self.faces)
    intersection = trimesh.intersections.mesh_plane(tri_polygons, plane.normal, plane.origin.coordinates)
    polys_1 = trimesh.intersections.slice_mesh_plane(tri_polygons, plane.opposite_normal, plane.origin.coordinates)
    polys_2 = trimesh.intersections.slice_mesh_plane(tri_polygons, plane.normal, plane.origin.coordinates)
    triangles_1 = []
    for triangle in polys_1.triangles:
      triangles_1.append(Polygon(triangle))
    polygon_1 = self._reshape(triangles_1)
    triangles_2 = []
    for triangle in polys_2.triangles:
      triangles_2.append(Polygon(triangle))
    polygon_2 = self._reshape(triangles_2)
    return polygon_1, polygon_2, intersection

  def _reshape(self, triangles) -> Polygon:
    edges_list = []
    for i in enumerate(triangles):
      for edge in triangles[i].edges:
        if not self._edge_in_edges_list(edge, edges_list):
          edges_list.append(edge)
        else:
          edges_list = self._remove_from_list(edge, edges_list)
    points = self._order_points(edges_list)
    return Polygon(points)

  @staticmethod
  def _edge_in_edges_list(edge, edges_list):
    for edge_element in edges_list:
      if (edge_element[0].distance_to_point(edge[0]) == 0 and edge_element[1].distance_to_point(edge[1]) == 0) or \
          (edge_element[1].distance_to_point(edge[0]) == 0 and edge_element[0].distance_to_point(
            edge[1]) == 0):
        return True
    return False

  @staticmethod
  def _order_points(edges_list):
    # todo: not sure that this method works for any case -> RECHECK
    points = edges_list[0]
    for _ in range(0, len(points)):
      for i in range(1, len(edges_list)):
        point_1 = edges_list[i][0]
        point_2 = points[len(points) - 1]
        if point_1.distance_to_point(point_2) == 0:
          points.append(edges_list[i][1])
    points.remove(points[len(points) - 1])
    array_points = []
    for point in points:
      array_points.append(point.coordinates)
    return np.array(array_points)

  @staticmethod
  def _remove_from_list(edge, edges_list):
    new_list = []
    for edge_element in edges_list:
      if not ((edge_element[0].distance_to_point(edge[0]) == 0 and edge_element[1].distance_to_point(
          edge[1]) == 0) or
              (edge_element[1].distance_to_point(edge[0]) == 0 and edge_element[0].distance_to_point(
                edge[1]) == 0)):
        new_list.append(edge_element)
    return new_list

  @property
  def vertices(self) -> np.ndarray:
    """
    Get polygon vertices
    :return: np.ndarray(int)
    """
    if self._vertices is None:
      vertices, self._vertices = [], []
      _ = [vertices.extend(s.coordinates) for s in self.triangles]
      for vertex_1 in vertices:
        found = False
        for vertex_2 in self._vertices:
          found = False
          power = 0
          for dimension in range(0, 3):
            power += math.pow(vertex_2[dimension] - vertex_1[dimension], 2)
          distance = math.sqrt(power)
          if distance == 0:
            found = True
            break
        if not found:
          self._vertices.append(vertex_1)
      self._vertices = np.asarray(self._vertices)
    return self._vertices

  @property
  def faces(self) -> List[List[int]]:
    """
    Get polygon triangular faces
    :return: [face]
    """
    if self._faces is None:
      self._faces = []

      for polygon in self.triangles:
        face = []
        points = polygon.coordinates
        if len(points) != 3:
          sub_polygons = polygon.triangles
          # todo: I modified this! To be checked @Guille
          if len(sub_polygons) < 1:
            continue
          for sub_polygon in sub_polygons:
            face = []
            points = sub_polygon.coordinates
            for point in points:
              face.append(self._position_of(point, face))
            self._faces.append(face)
        else:
          for point in points:
            face.append(self._position_of(point, face))
          self._faces.append(face)
    return self._faces

  def _position_of(self, point, face):
    """
    position of a specific point in the list of points that define a face
    :return: int
    """
    vertices = self.vertices
    for i in enumerate(vertices):
      # ensure not duplicated vertex
      power = 0
      vertex2 = vertices[i]
      for dimension in range(0, 3):
        power += math.pow(vertex2[dimension] - point[dimension], 2)
      distance = math.sqrt(power)
      if i not in face and distance == 0:
        return i
    return -1
